// This file is part of Selendra.

// Copyright (C) 2021-2022 Selendra.
// SPDX-License-Identifier: GPL-3.0-or-later WITH Classpath-exception-2.0

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

//! Mocks for the cdp engine module.

#![cfg(test)]

use super::*;
use frame_support::{
	construct_runtime, ord_parameter_types, parameter_types,
	traits::{ConstU128, ConstU32, ConstU64, Everything, Nothing},
	PalletId,
};
use frame_system::EnsureSignedBy;
use orml_traits::parameter_type_with_key;
use primitives::{DexShare, Moment, TokenSymbol, TradingPair};
use sp_core::{crypto::AccountId32, H256};
use sp_runtime::{
	testing::{Header, TestXt},
	traits::{AccountIdConversion, IdentityLookup, One as OneT},
};
use sp_std::{cell::RefCell, str::FromStr};
use support::{mocks::MockStableAsset, AuctionManager, EmergencyShutdown, SpecificJointsSwap};

pub type AccountId = AccountId32;
pub type BlockNumber = u64;
pub type AuctionId = u32;

pub const ALICE: AccountId = AccountId32::new([1u8; 32]);
pub const BOB: AccountId = AccountId32::new([2u8; 32]);
pub const CAROL: AccountId = AccountId32::new([3u8; 32]);
pub const SEL: CurrencyId = CurrencyId::Token(TokenSymbol::SEL);
pub const KUSD: CurrencyId = CurrencyId::Token(TokenSymbol::KUSD);
pub const BTC: CurrencyId = CurrencyId::Token(TokenSymbol::RENBTC);
pub const DOT: CurrencyId = CurrencyId::Token(TokenSymbol::DOT);
pub const LP_KUSD_DOT: CurrencyId =
	CurrencyId::DexShare(DexShare::Token(TokenSymbol::KUSD), DexShare::Token(TokenSymbol::DOT));
pub const LP_DOT_BTC: CurrencyId =
	CurrencyId::DexShare(DexShare::Token(TokenSymbol::RENBTC), DexShare::Token(TokenSymbol::DOT));

mod cdp_engine {
	pub use super::super::*;
}

impl frame_system::Config for Runtime {
	type Origin = Origin;
	type Index = u64;
	type BlockNumber = BlockNumber;
	type Call = Call;
	type Hash = H256;
	type Hashing = ::sp_runtime::traits::BlakeTwo256;
	type AccountId = AccountId;
	type Lookup = IdentityLookup<Self::AccountId>;
	type Header = Header;
	type Event = Event;
	type BlockHashCount = ConstU64<250>;
	type BlockWeights = ();
	type BlockLength = ();
	type Version = ();
	type PalletInfo = PalletInfo;
	type AccountData = pallet_balances::AccountData<Balance>;
	type OnNewAccount = ();
	type OnKilledAccount = ();
	type DbWeight = ();
	type BaseCallFilter = Everything;
	type SystemWeightInfo = ();
	type SS58Prefix = ();
	type OnSetCode = ();
	type MaxConsumers = ConstU32<16>;
}

parameter_type_with_key! {
	pub ExistentialDeposits: |_currency_id: CurrencyId| -> Balance {
		Default::default()
	};
}

impl orml_tokens::Config for Runtime {
	type Event = Event;
	type Balance = Balance;
	type Amount = Amount;
	type CurrencyId = CurrencyId;
	type WeightInfo = ();
	type ExistentialDeposits = ExistentialDeposits;
	type OnDust = ();
	type MaxLocks = ();
	type MaxReserves = ();
	type ReserveIdentifier = [u8; 8];
	type DustRemovalWhitelist = Nothing;
	type OnNewTokenAccount = ();
	type OnKilledTokenAccount = ();
}

impl pallet_balances::Config for Runtime {
	type Balance = Balance;
	type DustRemoval = ();
	type Event = Event;
	type ExistentialDeposit = ConstU128<1>;
	type AccountStore = frame_system::Pallet<Runtime>;
	type MaxLocks = ();
	type MaxReserves = ();
	type ReserveIdentifier = [u8; 8];
	type WeightInfo = ();
}
pub type AdaptedBasicCurrency =
	orml_currencies::BasicCurrencyAdapter<Runtime, PalletBalances, Amount, BlockNumber>;

parameter_types! {
	pub const GetNativeCurrencyId: CurrencyId = SEL;
}

impl orml_currencies::Config for Runtime {
	type MultiCurrency = Tokens;
	type NativeCurrency = AdaptedBasicCurrency;
	type GetNativeCurrencyId = GetNativeCurrencyId;
	type WeightInfo = ();
}

parameter_types! {
	pub const LoansPalletId: PalletId = PalletId(*b"sel/loan");
}

impl loans::Config for Runtime {
	type Event = Event;
	type Currency = Currencies;
	type RiskManager = CDPEngineModule;
	type CDPTreasury = CDPTreasuryModule;
	type PalletId = LoansPalletId;
	type OnUpdateLoan = ();
}

thread_local! {
	static BTC_PRICE: RefCell<Option<Price>> = RefCell::new(Some(Price::one()));
	static DOT_PRICE: RefCell<Option<Price>> = RefCell::new(Some(Price::one()));
	static LP_KUSD_DOT_PRICE: RefCell<Option<Price>> = RefCell::new(Some(Price::one()));
	static LP_DOT_BTC_PRICE: RefCell<Option<Price>> = RefCell::new(Some(Price::one()));
}

pub struct MockPriceSource;
impl MockPriceSource {
	pub fn set_price(currency_id: CurrencyId, price: Option<Price>) {
		match currency_id {
			BTC => BTC_PRICE.with(|v| *v.borrow_mut() = price),
			DOT => DOT_PRICE.with(|v| *v.borrow_mut() = price),
			LP_KUSD_DOT => LP_KUSD_DOT_PRICE.with(|v| *v.borrow_mut() = price),
			LP_DOT_BTC => LP_DOT_BTC_PRICE.with(|v| *v.borrow_mut() = price),
			_ => {},
		}
	}
}
impl PriceProvider<CurrencyId> for MockPriceSource {
	fn get_price(currency_id: CurrencyId) -> Option<Price> {
		match currency_id {
			BTC => BTC_PRICE.with(|v| *v.borrow()),
			DOT => DOT_PRICE.with(|v| *v.borrow()),
			KUSD => Some(Price::one()),
			LP_KUSD_DOT => LP_KUSD_DOT_PRICE.with(|v| *v.borrow()),
			LP_DOT_BTC => LP_DOT_BTC_PRICE.with(|v| *v.borrow()),
			_ => None,
		}
	}
}

thread_local! {
	pub static AUCTION: RefCell<Option<(AccountId, CurrencyId, Balance, Balance)>> = RefCell::new(None);
}

pub struct MockAuctionManager;
impl MockAuctionManager {
	pub fn auction() -> Option<(AccountId, CurrencyId, Balance, Balance)> {
		AUCTION.with(|v| {
			let cloned = v.borrow().clone();
			cloned
		})
	}
}
impl AuctionManager<AccountId> for MockAuctionManager {
	type Balance = Balance;
	type CurrencyId = CurrencyId;
	type AuctionId = AuctionId;

	fn new_collateral_auction(
		refund_recipient: &AccountId,
		currency_id: Self::CurrencyId,
		amount: Self::Balance,
		target: Self::Balance,
	) -> DispatchResult {
		AUCTION.with(|v| {
			*v.borrow_mut() = Some((refund_recipient.clone(), currency_id, amount, target))
		});
		Ok(())
	}

	fn cancel_auction(_id: Self::AuctionId) -> DispatchResult {
		AUCTION.with(|v| *v.borrow_mut() = None);
		Ok(())
	}

	fn get_total_target_in_auction() -> Self::Balance {
		Self::auction().map(|auction| auction.3).unwrap_or_default()
	}

	fn get_total_collateral_in_auction(_id: Self::CurrencyId) -> Self::Balance {
		Self::auction().map(|auction| auction.2).unwrap_or_default()
	}
}

parameter_types! {
	pub const GetStableCurrencyId: CurrencyId = KUSD;
	pub const CDPTreasuryPalletId: PalletId = PalletId(*b"sel/cdpt");
	pub TreasuryAccount: AccountId = PalletId(*b"sel/hztr").into_account_truncating();
	pub AlternativeSwapPathJointList: Vec<Vec<CurrencyId>> = vec![
		vec![SEL],
	];
}

impl cdp_treasury::Config for Runtime {
	type Event = Event;
	type Currency = Currencies;
	type GetStableCurrencyId = GetStableCurrencyId;
	type AuctionManagerHandler = MockAuctionManager;
	type UpdateOrigin = EnsureSignedBy<One, AccountId>;
	type DEX = DEXModule;
	type Swap = SpecificJointsSwap<DEXModule, AlternativeSwapPathJointList>;
	type MaxAuctionsCount = ConstU32<10_000>;
	type PalletId = CDPTreasuryPalletId;
	type TreasuryAccount = TreasuryAccount;
	type WeightInfo = ();
	type StableAsset = MockStableAsset<CurrencyId, Balance, AccountId, BlockNumber>;
}

parameter_types! {
	pub const DEXPalletId: PalletId = PalletId(*b"sel/dexm");
	pub const GetExchangeFee: (u32, u32) = (0, 100);
	pub EnabledTradingPairs: Vec<TradingPair> = vec![
		TradingPair::from_currency_ids(KUSD, BTC).unwrap(),
		TradingPair::from_currency_ids(KUSD, DOT).unwrap(),
		TradingPair::from_currency_ids(SEL, BTC).unwrap(),
		TradingPair::from_currency_ids(SEL, DOT).unwrap(),
		TradingPair::from_currency_ids(SEL, KUSD).unwrap(),
	];
}

impl dex::Config for Runtime {
	type Event = Event;
	type Currency = Currencies;
	type GetExchangeFee = GetExchangeFee;
	type TradingPathLimit = ConstU32<4>;
	type PalletId = DEXPalletId;
	type Erc20InfoMapping = ();
	type DEXIncentives = ();
	type WeightInfo = ();
	type ListingOrigin = EnsureSignedBy<One, AccountId>;
	type ExtendedProvisioningBlocks = ConstU64<0>;
	type OnLiquidityPoolUpdated = ();
}

impl pallet_timestamp::Config for Runtime {
	type Moment = Moment;
	type OnTimestampSet = ();
	type MinimumPeriod = ConstU64<1_000>;
	type WeightInfo = ();
}

impl evm_accounts::Config for Runtime {
	type Event = Event;
	type Currency = PalletBalances;
	type ChainId = ();
	type AddressMapping = evm_accounts::EvmAddressMapping<Runtime>;
	type TransferAll = Currencies;
	type WeightInfo = ();
}

thread_local! {
	static IS_SHUTDOWN: RefCell<bool> = RefCell::new(false);
}

pub fn mock_shutdown() {
	IS_SHUTDOWN.with(|v| *v.borrow_mut() = true)
}

pub fn liquidation_contract_addr() -> EvmAddress {
	EvmAddress::from_str(&"0x1000000000000000000000000000000000000000").unwrap()
}

pub struct MockEmergencyShutdown;
impl EmergencyShutdown for MockEmergencyShutdown {
	fn is_shutdown() -> bool {
		IS_SHUTDOWN.with(|v| *v.borrow_mut())
	}
}

thread_local! {
	static LIQUIDATED: RefCell<(EvmAddress, EvmAddress, Balance, Balance)> = RefCell::new((EvmAddress::default(), EvmAddress::default(), 0, 0));
	static TRANSFERRED: RefCell<(EvmAddress, Balance)> = RefCell::new((EvmAddress::default(), 0));
	static REFUNDED: RefCell<(EvmAddress, Balance)> = RefCell::new((EvmAddress::default(), 0));
	static LIQUIDATION_RESULT: RefCell<DispatchResult> = RefCell::new(Err(Error::<Runtime>::LiquidationFailed.into()));
	static REPAYMENT: RefCell<Option<Balance>> = RefCell::new(None);
}

pub struct MockLiquidationEvmBridge;
impl MockLiquidationEvmBridge {
	pub fn liquidated() -> (EvmAddress, EvmAddress, Balance, Balance) {
		LIQUIDATED.with(|v| v.borrow().clone())
	}
	pub fn transferred() -> (EvmAddress, Balance) {
		TRANSFERRED.with(|v| v.borrow().clone())
	}
	pub fn refunded() -> (EvmAddress, Balance) {
		REFUNDED.with(|v| v.borrow().clone())
	}
	pub fn reset() {
		LIQUIDATION_RESULT
			.with(|v| *v.borrow_mut() = Err(Error::<Runtime>::LiquidationFailed.into()));
		REPAYMENT.with(|v| *v.borrow_mut() = None);
	}
	pub fn set_liquidation_result(r: DispatchResult) {
		LIQUIDATION_RESULT.with(|v| *v.borrow_mut() = r);
	}
	pub fn set_repayment(repayment: Balance) {
		REPAYMENT.with(|v| *v.borrow_mut() = Some(repayment));
	}
}
impl LiquidationEvmBridge for MockLiquidationEvmBridge {
	fn liquidate(
		_context: InvokeContext,
		collateral: EvmAddress,
		repay_dest: EvmAddress,
		amount: Balance,
		min_repayment: Balance,
	) -> DispatchResult {
		let result = LIQUIDATION_RESULT.with(|v| v.borrow().clone());
		if result.is_ok() {
			let repayment = if let Some(r) = REPAYMENT.with(|v| v.borrow().clone()) {
				r
			} else {
				min_repayment
			};
			let _ = Currencies::deposit(
				GetStableCurrencyId::get(),
				&CDPEngineModule::account_id(),
				repayment,
			);
		}
		LIQUIDATED.with(|v| *v.borrow_mut() = (collateral, repay_dest, amount, min_repayment));
		result
	}
	fn on_collateral_transfer(_context: InvokeContext, collateral: EvmAddress, amount: Balance) {
		TRANSFERRED.with(|v| *v.borrow_mut() = (collateral, amount));
	}
	fn on_repayment_refund(_context: InvokeContext, collateral: EvmAddress, repayment: Balance) {
		REFUNDED.with(|v| *v.borrow_mut() = (collateral, repayment));
	}
}

ord_parameter_types! {
	pub const One: AccountId = ALICE;
}

parameter_type_with_key! {
	pub MinimumCollateralAmount: |_currency_id: CurrencyId| -> Balance {
		10
	};
}

parameter_types! {
	pub DefaultLiquidationRatio: Ratio = Ratio::saturating_from_rational(3, 2);
	pub DefaultDebitExchangeRate: ExchangeRate = ExchangeRate::saturating_from_rational(1, 10);
	pub DefaultLiquidationPenalty: Rate = Rate::saturating_from_rational(10, 100);
	pub MaxSwapSlippageCompareToOracle: Ratio = Ratio::saturating_from_rational(50, 100);
	pub MaxLiquidationContractSlippage: Ratio = Ratio::saturating_from_rational(80, 100);
	pub const CDPEnginePalletId: PalletId = PalletId(*b"sel/cdpe");
}

impl Config for Runtime {
	type Event = Event;
	type PriceSource = MockPriceSource;
	type DefaultLiquidationRatio = DefaultLiquidationRatio;
	type DefaultDebitExchangeRate = DefaultDebitExchangeRate;
	type DefaultLiquidationPenalty = DefaultLiquidationPenalty;
	type MinimumDebitValue = ConstU128<2>;
	type MinimumCollateralAmount = MinimumCollateralAmount;
	type GetStableCurrencyId = GetStableCurrencyId;
	type CDPTreasury = CDPTreasuryModule;
	type UpdateOrigin = EnsureSignedBy<One, AccountId>;
	type MaxSwapSlippageCompareToOracle = MaxSwapSlippageCompareToOracle;
	type UnsignedPriority = ConstU64<1048576>; // 1 << 20
	type EmergencyShutdown = MockEmergencyShutdown;
	type UnixTime = Timestamp;
	type Currency = Currencies;
	type DEX = DEXModule;
	type LiquidationContractsUpdateOrigin = EnsureSignedBy<One, AccountId>;
	type MaxLiquidationContractSlippage = MaxLiquidationContractSlippage;
	type MaxLiquidationContracts = ConstU32<10>;
	type LiquidationEvmBridge = MockLiquidationEvmBridge;
	type PalletId = CDPEnginePalletId;
	type EvmAddressMapping = evm_accounts::EvmAddressMapping<Runtime>;
	type Swap = SpecificJointsSwap<DEXModule, AlternativeSwapPathJointList>;
	type WeightInfo = ();
}

type UncheckedExtrinsic = frame_system::mocking::MockUncheckedExtrinsic<Runtime>;
type Block = frame_system::mocking::MockBlock<Runtime>;

construct_runtime!(
	pub enum Runtime where
		Block = Block,
		NodeBlock = Block,
		UncheckedExtrinsic = UncheckedExtrinsic,
	{
		System: frame_system::{Pallet, Call, Storage, Config, Event<T>},
		CDPEngineModule: cdp_engine::{Pallet, Storage, Call, Event<T>, Config, ValidateUnsigned},
		CDPTreasuryModule: cdp_treasury::{Pallet, Storage, Call, Config, Event<T>},
		Currencies: orml_currencies::{Pallet, Call},
		Tokens: orml_tokens::{Pallet, Storage, Event<T>, Config<T>},
		LoansModule: loans::{Pallet, Storage, Call, Event<T>},
		PalletBalances: pallet_balances::{Pallet, Call, Storage, Event<T>},
		DEXModule: dex::{Pallet, Storage, Call, Event<T>, Config<T>},
		Timestamp: pallet_timestamp::{Pallet, Call, Storage, Inherent},
		EvmAccounts: evm_accounts::{Pallet, Call, Storage, Event<T>},
	}
);

/// An extrinsic type used for tests.
pub type Extrinsic = TestXt<Call, ()>;

impl<LocalCall> SendTransactionTypes<LocalCall> for Runtime
where
	Call: From<LocalCall>,
{
	type OverarchingCall = Call;
	type Extrinsic = Extrinsic;
}

pub struct ExtBuilder {
	balances: Vec<(AccountId, CurrencyId, Balance)>,
}

impl Default for ExtBuilder {
	fn default() -> Self {
		Self {
			balances: vec![
				(ALICE, BTC, 1000),
				(BOB, BTC, 1000),
				(CAROL, BTC, 10000),
				(ALICE, DOT, 1000),
				(BOB, DOT, 1000),
				(CAROL, DOT, 10000),
				(CAROL, KUSD, 10000),
			],
		}
	}
}

impl ExtBuilder {
	pub fn build(self) -> sp_io::TestExternalities {
		let mut t = frame_system::GenesisConfig::default().build_storage::<Runtime>().unwrap();

		pallet_balances::GenesisConfig::<Runtime> { balances: vec![(CAROL, 10000)] }
			.assimilate_storage(&mut t)
			.unwrap();

		orml_tokens::GenesisConfig::<Runtime> { balances: self.balances }
			.assimilate_storage(&mut t)
			.unwrap();

		dex::GenesisConfig::<Runtime> {
			initial_listing_trading_pairs: vec![],
			initial_enabled_trading_pairs: EnabledTradingPairs::get(),
			initial_added_liquidity_pools: vec![],
		}
		.assimilate_storage(&mut t)
		.unwrap();

		MockLiquidationEvmBridge::reset();

		t.into()
	}

	pub fn lots_of_accounts() -> Self {
		let mut balances = Vec::new();
		for i in 0..1001u32 {
			balances.push((account_id_from_u32(i), BTC, 1000));
		}
		Self { balances }
	}
}

pub fn account_id_from_u32(num: u32) -> AccountId {
	let mut data = [0u8; 32];
	let index = num.to_le_bytes();
	data[0..4].copy_from_slice(&index[..]);
	AccountId::new(data)
}
